#!/bin/bash
################################################################################
#
#  build_kernel.sh
#
#  Copyright (c) 2019-2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
################################################################################

function usage {
    echo "Usage: ${BASH_SOURCE[0]} path_to_platform_tar|path_to_platform_dir output_folder" 1>&2
    exit 1
}

################################################################################
#
#  I N P U T
#
################################################################################
# Folder for platform tarball.
PLATFORM_TARBALL="${1}"

# Target directory for output artifacts.
TARGET_DIR="${2}/${TARGET_DEVICE}/busybox"

################################################################################
#
#  V A R I A B L E S
#
################################################################################

# Retrieve the directory where the script is currently held
SCRIPT_BASE_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"

# Configuration file for the build.
CONFIG_FILE="${SCRIPT_BASE_DIR}/build_busybox_config.sh"

# Workspace directory & relevant temp folders.
if [ -d "${PLATFORM_TARBALL}" ]; then
    WORKSPACE_DIR="${PLATFORM_TARBALL}"
else
    WORKSPACE_DIR="$(mktemp -d)"
fi

TOOLCHAIN_DIR="${WORKSPACE_DIR}/toolchain"
PLATFORM_EXTRACT_DIR="${WORKSPACE_DIR}/src"
WORKSPACE_OUT_DIR="${WORKSPACE_DIR}/out"
BUSYBOX_OUT_DIR="${WORKSPACE_OUT_DIR}/busybox"

for d in "${TOOLCHAIN_DIR}" "${PLATFORM_EXTRACT_DIR}" "$BUSYBOX_OUT_DIR"
do
    mkdir -p "${d}"
done

# Remove workspace directory upon completion.
trap "rm -rf $WORKSPACE_DIR" EXIT

PARALLEL_EXECUTION="-j8"

function validate_input_params {
    if [[ ! -f "${PLATFORM_TARBALL}" ]] && [[ ! -d "${PLATFORM_TARBALL}" ]]
    then
        echo "ERROR: Platform tarball or directory not found."
        usage
    fi

    if [[ ! -f "${CONFIG_FILE}" ]]
    then
        echo "ERROR: Could not find config file ${CONFIG_FILE}. Please check" \
             "that you have extracted the build script properly and try again."
        usage
    fi
}

function display_config {
    echo "-------------------------------------------------------------------------"
    echo "PLATFORM SOURCE: ${PLATFORM_TARBALL}"
    echo "TARGET DIRECTORY: ${TARGET_DIR}"
    echo "BUSYBOX SUBPATH: ${BUSYBOX_SUBPATH}"
    echo "BUSYBOX CONFIG: ${BUSYBOX_DEFCONFIG_NAME}"
    echo "TARGET ARCHITECTURE: ${TARGET_ARCH}"
    echo "TOOLCHAIN REPO: ${TOOLCHAIN_REPO}"
    echo "TOOLCHAIN PREFIX: ${TOOLCHAIN_PREFIX}"
    echo "-------------------------------------------------------------------------"
    echo "Sleeping 3 seconds before continuing."
    sleep 3
}

function setup_output_dir {
    if [[ -d "${TARGET_DIR}" ]]
    then
        FILECOUNT=$(find "${TARGET_DIR}" -type f | wc -l)
        if [[ ${FILECOUNT} -gt 0 ]]
        then
            echo "ERROR: Destination folder is not empty. Refusing to build" \
                 "to a non-clean target"
            exit 3
        fi
    else
        echo "Making target directory ${TARGET_DIR}"
        mkdir -p "${TARGET_DIR}"

        if [[ $? -ne 0 ]]
        then
            echo "ERROR: Could not make target directory ${TARGET_DIR}"
            exit 1
        fi
    fi
}

function download_toolchain {
    echo "Downloading toolchain ${TOOLCHAIN_REPO} to ${TOOLCHAIN_DIR}"
    curl -L ${TOOLCHAIN_REPO} | tar -xJ -C ${TOOLCHAIN_DIR}
    if [[ $? -ne 0 ]]
    then
        echo "ERROR: Could not download toolchain from ${TOOLCHAIN_REPO}."
        exit 2
    fi
}

function extract_tarball {
    echo "Extracting tarball to ${PLATFORM_EXTRACT_DIR}"
    tar xf "${PLATFORM_TARBALL}" -C ${PLATFORM_EXTRACT_DIR}
}

function exec_build_busybox {
    CCOMPILE="${TOOLCHAIN_DIR}/${TOOLCHAIN_NAME}/bin/${TOOLCHAIN_PREFIX}"

    if [[ -n "${BUSYBOX_SUBPATH}" ]]
    then
        MAKE_ARGS="-C ${BUSYBOX_SUBPATH}"
    fi

    MAKE_ARGS="${MAKE_ARGS} O=${BUSYBOX_OUT_DIR} ARCH=${TARGET_ARCH} CROSS_COMPILE=${CCOMPILE}"
    echo "Base make args: ${MAKE_ARGS}"

    # Move into the build base folder.
    pushd "${PLATFORM_EXTRACT_DIR}"

    # Step 1: defconfig
    echo "Make defconfig: make ${MAKE_ARGS} ${BUSYBOX_DEFCONFIG_NAME}"
    make ${MAKE_ARGS} ${BUSYBOX_DEFCONFIG_NAME}

    # Step 2: output config, for reference
    local OUTPUT_CFG="${BUSYBOX_OUT_DIR}/.config"
    echo ".config contents"
    echo "---------------------------------------------------------------------"
    cat "${OUTPUT_CFG}"
    echo "---------------------------------------------------------------------"

    # Step 3: full make
    make ${MAKE_ARGS} ${PARALLEL_EXECUTION} 

    if [[ $? != 0 ]]; then
        echo "ERROR: Failed to build busybox" >&2
        exit 1
    fi

    popd
}

function copy_to_output {
    local IFS=":"
    echo "Copying Busybox files to output"
    pushd "${BUSYBOX_OUT_DIR}"
    mkdir "${TARGET_DIR}/busybox"
    for IMAGE in ${BUSYBOX_IMAGES};do
        local BASEDIR="$(dirname "${IMAGE}")"
        if [[ ! -d "${TARGET_DIR}/busybox/${BASEDIR}" ]]
        then
            mkdir -p "${TARGET_DIR}/busybox/${BASEDIR}"
        fi
        cp -v ${IMAGE} "${TARGET_DIR}/busybox/${IMAGE}"
    done
    popd
}

function validate_output {
    echo "Listing output files"
    local IFS=":"
    for IMAGE in ${BUSYBOX_IMAGES};do
        if [ ! -f ${TARGET_DIR}/busybox/${IMAGE} ]; then
            echo "ERROR: Missing busybox output image ${IMAGE}" >&2
            exit 1
        fi
        ls -l ${TARGET_DIR}/busybox/${IMAGE}
    done
}

################################################################################
#
#  M A I N
#
################################################################################

# Phase 1: Set up execution
validate_input_params
source "${CONFIG_FILE}"
setup_output_dir
TARGET_DIR="$(cd "${TARGET_DIR}" && pwd)"
display_config

# Phase 2: Set up environment
if [ -n "${TOOLCHAIN_NAME}" ] && [ -n "$(ls -A /tmp/${TOOLCHAIN_NAME})" ]; then
    TOOLCHAIN_DIR="/tmp"
else
    download_toolchain
fi
extract_tarball

# Phase 3: build busybox
exec_build_busybox

# Phase 4: move to output
copy_to_output

# Phase 5: verify output
validate_output
