/*
 * Copyright (c) 2019-2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * This file defines the interface between Red and Blue. Included in multiple builds
 */

#ifndef __COCOA_BLUE_UART_BOOT_H__
#define __COCOA_BLUE_UART_BOOT_H__

#define COCOA_BLUE_UART_BOOT_STATIC_ASSERT(COND,MSG) \
    typedef char static_assertion_##MSG[(!!(COND))*2-1]

#define COCOA_BLUE_UART_VERSION     7

typedef enum _UART_Boot_IR_Leds {
    UB_IR_LED_0 = 0,
    UB_IR_LED_10,
    UB_IR_LED_20,
    UB_IR_LED_50
} UART_Boot_IR_Leds;

typedef enum _UART_Boot_Night_Mode {
    UB_Day_Mode = 0,
    UB_Night_Mode,
    UB_Night_Colour_Mode,
    UB_Hdr_Mode,            // added in COCOA-12644 [abc123] [SAV533] Support SOI JX-F51 image sensor
    UB_Wled_On_Mode,        // added in COCOA-12482 [abc123] Separate WLED tuning from daymode tuning
    UB_Wled_After_Night_Mode, // added in COCOA-14916: [abc123] Use the previous isp mode for wled mode
                              // an isp tuning file of WLED_ON mode is used for this mode.
} UART_Boot_Night_Mode;

typedef enum _UART_Boot_Fps {
    UB_FPS30 = 0,
    UB_FPS24,
    UB_FPS15,
    UB_FPS8,
} UART_Boot_Fps;

typedef enum _UART_Boot_BitRate {
    UB_BitRate_2m = 0,
    UB_BitRate_1500k,
    UB_BitRate_1000k,
    UB_BitRate_750k,
    UB_BitRate_500k,
    UB_BitRate_400k,
    UB_BitRate_350k,
    UB_BitRate_300k,
} UART_Boot_BitRate;

typedef enum _UART_Boot_Resolution {
    // 16:9
    UB_Res_1080P = 0,
    UB_Res_720P,
    UB_Res_360P,

    // 1:1
    /* UART_Boot_Resolution uses 2 bits in section2.
       To maintain backward compatibility, use the same
       enum for 1:1 resolutions as well but starting with 0 */
    UB_Res_1536x1536P = 0,
    UB_Res_1080x1080P,
    UB_Res_480x480P
} UART_Boot_Resolution;

typedef enum _UART_Boot_Flicker {
    UB_FLICKER_DISABLE = 0,
    UB_FLICKER_60HZ = 1,
    UB_FLICKER_50HZ = 2,
    UB_FLICKER_AUTO = 3
} UART_Boot_Flicker;

typedef enum _UART_Boot_VideoCodec {
    UB_VIDEO_CODEC_H264 = 0,
    UB_VIDEO_CODEC_H265 = 1
} UART_Boot_VideoCodec;

typedef enum _UART_Boot_Reason {
    UB_Reason_Streaming = 0,
    UB_Reason_Motion = 1,
    UB_Reason_Earcon = 2,
    UB_Reason_Snapshot = 3
} UART_Boot_Reason;

#define UART_BOOT_CHECKSUM_SIZE     (8)
#define UART_BOOT_CHECKSUM_MASK     ( (1 << UART_BOOT_CHECKSUM_SIZE) - 1)

typedef enum _UART_Boot_SectionNum {
    UB_Section1 = 1,
    UB_Section2,
    UB_SectionMax
} UART_Boot_SectionNum;

typedef struct _uart_boot_section1 {
    unsigned int            ProtocolVersion : 8;
    unsigned int            SensorGain : 16;    // [1024, 32*1024] equivalent to 1x..32x gain
    unsigned int            IspGain : 18;       // [1024, (256*1024-1)] equivalent to 1x..255.9999x gain
    unsigned int            ExposureTime : 18;  // [1, 200000] us
    unsigned int            RedGain : 13;       // [1024, (8*1024)-1]
    unsigned int            BlueGain : 13;      // [1024, (8*1024)-1]
    unsigned int            GreenGain : 13;     // [1024, (8*1024)-1]
    UART_Boot_IR_Leds       IRLeds : 2;
    UART_Boot_Night_Mode    DayNightMode : 3;
    UART_Boot_Fps           SensorFrameRate : 2;
    UART_Boot_Flicker       AntiFlicker : 2;
    unsigned int            RlmdHd : 1;
    unsigned int            BitSum : UART_BOOT_CHECKSUM_SIZE;
} __attribute__((packed)) uart_boot_section1;

// Changes in size will break backward-compatibility and need an update
// to the protocol version
COCOA_BLUE_UART_BOOT_STATIC_ASSERT(sizeof(uart_boot_section1) == 15, uart_boot_section1_size);

typedef struct _uart_boot_section2 {
    unsigned int            TimeSeconds : 32;
    unsigned int            TimeMilliseconds : 16;
    unsigned int            BitSum : UART_BOOT_CHECKSUM_SIZE;
    UART_Boot_Reason        BootReason:2;
    UART_Boot_Resolution    Resolution : 2;
    UART_Boot_Fps           EncFrameRate : 2;
    unsigned int            RegionsNum : 2;
    UART_Boot_BitRate       TargetBitRate : 3;

    // Regions given as ratio of total image size, with range 0-1023
    // 1023 represents 1.0 * image size on each axis
    unsigned int            region0Xratio : 10;
    unsigned int            region0Yratio : 10;
    unsigned int            region0Wratio : 10;
    unsigned int            region0Hratio : 10;

    unsigned int            region1Xratio : 10;
    unsigned int            region1Yratio : 10;
    unsigned int            region1Wratio : 10;
    unsigned int            region1Hratio : 10;

    unsigned int            RMSEnabled : 1;
    unsigned int            HighResYuv : 1;
    UART_Boot_VideoCodec    VideoCodec : 2;
    unsigned int            EarlyPrepareDisabled : 1;
} __attribute__((packed)) uart_boot_section2;

// Changes in size will break backward-compatibility and need an update
// to the protocol version
COCOA_BLUE_UART_BOOT_STATIC_ASSERT(sizeof(uart_boot_section2) == 19, uart_boot_section2_size);

#endif //__COCOA_BLUE_UART_BOOT_H__
